<?php

/**
 * Class GovPay.
 */
class GovPay {
  protected $baseUrl;
  protected $apiKey;

  protected $lastError;

  /**
   * GovPay constructor.
   */
  public function __construct() {
    if (variable_get('gov_pay__apikey', NULL)) {
      $this->apiKey = variable_get('gov_pay__apikey', NULL);
      $this->baseUrl = "https://publicapi.payments.service.gov.uk";
    }
  }

  /**
   * Generate a UUID.
   *
   * (Lifted from Drupal uuid module).
   */
  public static function uuid() {
    // We limit each generated number to 16 bits (maximum value 0xffff)
    // because mt_rand() returns a *signed* integer, and hence a 32-bit
    // value can only have a 31-bit magnitude. Constructing a 32-bit
    // number from two 16-bit random numbers guarantees that all 32 bits
    // are random.
    return sprintf('%04x%04x-%04x-4%03x-%04x-%04x%04x%04x',
      // 32 bits for "time_low".
      mt_rand(0, 0xffff), mt_rand(0, 0xffff),
      // 16 bits for "time_mid".
      mt_rand(0, 0xffff),
      // 12 bits after the initial 0100 (version 4) for "time_hi_and_version".
      mt_rand(0, 0x0fff),
      // 16 bits in total for "clk_seq_hi_res" and "clk_seq_low", with the
      // most significant 2 bits of clk_seq_hi_res set to '10'. We do a
      // bitwise OR of a random 14-bit value (maximum 0x3fff) with 0x8000
      // (a 16-bit integer with only the most significant bit set).
      mt_rand(0, 0x3fff) | 0x8000,
      // 48 bits for "node".
      mt_rand(0, 0xffff), mt_rand(0, 0xffff), mt_rand(0, 0xffff)
    );
  }

  /**
   * Get payment via GovPay paymentId.
   */
  public function getPayment($paymentId) {

    // Check api key has been configured.
    if (!isset($this->apiKey)) {
      return;
    }

    $url = "$this->baseUrl/v1/payments/$paymentId";
    // Instantiate request.
    $headers = [
      'CONTENT-TYPE' => "application/json",
      'Authorization' => "Bearer $this->apiKey",
    ];
    // Make request.
    $options = [
      'headers' => $headers,
      'method' => 'GET',
    ];
    $result = drupal_http_request($url, $options);
    // If request returns data, decode.
    $resultData = json_decode($result->data);
    $resultStatus = $result->code;

    // If good request, process.
    if ($resultStatus === '200') {
      return $resultData;
    }
  }

  /**
   * Create a Payment in GovPay.
   *
   * @param int $amount
   *   Amount in pence for payment.
   * @param string $description
   *   Payment description shown to user in GovPay.
   * @param array $extras
   *   Additional parts to save (Node, Submission ID).
   *
   * @return mixed|null
   *   Return the GovPay Payment Object, or NULL if an issue occurs.
   */
  public function makePayment($amount, $description, $extras = []) {

    // Check api key has been configured.
    if (!isset($this->apiKey)) {
      $this->lastError = "No API key configured";
      return FALSE;
    }

    global $base_url;
    $sid = isset($extras['sid']) ? $extras['sid'] : NULL;
    $nid = isset($extras['nid']) ? $extras['nid'] : NULL;
    $uuid = isset($extras['uuid']) ? $extras['uuid'] : self::uuid();
    $reference = !empty($extras['reference']) ? $extras['reference'] : 'n/a';

    $returnUrl = "$base_url/gov-pay-confirmation/$uuid";
    if (!is_null($nid)) {
      $returnUrl .= "/$nid";
      if (!is_null($sid)) {
        $returnUrl .= "/$sid";
      }
    }

    $url = "$this->baseUrl/v1/payments";
    // Instantiate request.
    $headers = [
      'CONTENT-TYPE' => "application/json",
      'Authorization' => "Bearer $this->apiKey",
    ];
    $reqBody = [
      "amount" => $amount,
      "reference" => $reference,
      "return_url" => $returnUrl,
      "description" => $description,
      "language" => "en",
    ];
    $jsonBody = json_encode($reqBody, JSON_UNESCAPED_SLASHES);
    // Make request.
    $options = [
      'headers' => $headers,
      'data' => $jsonBody,
      'method' => 'POST',
    ];
    $result = drupal_http_request($url, $options);

    $resultData = (object) [];
    if (!empty($result->data)) {
      $resultData = json_decode($result->data);
    }

    // If good request, process.
    if ($result->code === '201') {
      $return = $resultData;

      // Setup entity record.
      $payment = entity_get_controller('gov_payment')->create();
      $payment->payment_id = $resultData->payment_id;
      $payment->uuid = $uuid;
      if (!is_null($nid)) {
        $payment->nid = $nid;
      }
      if (!is_null($sid)) {
        $payment->sid = $sid;
      }
      $payment->status = $resultData->state->status;
      $payment->amount = $amount;
      $payment->date_created = strtotime('now');
      $payment->date_modified = strtotime('now');

      // Return save outcome.
      $return->local_save = entity_save('gov_payment', $payment);
      return $return;
    }
    else {
      // Decode error messages if present.
      if (!empty($resultData->code)) {
        $this->lastError = $resultData->code;
      }

      $variables = [
        '@code' => !empty($resultData->code) ? $resultData->code : '',
        '@description' => !empty($resultData->description) ? $resultData->description : '',
      ];

      watchdog('gov_uk_pay',"Error creating payment - @code, @description",
        $variables, WATCHDOG_ERROR);
    }
  }

  /**
   * Retrieve onsite record of a gov_payment.
   *
   * @param string $uuid
   *   UUID of payment.
   * @param int $nid
   *   Optional NID of payment.
   * @param int $sid
   *   Optional SID of payment.
   *
   * @return array
   *   Return loaded gov_payment entities.
   */
  public function fetchGovPayment($uuid, $nid = NULL, $sid = NULL) {

    // Initial Query.
    $query = new \EntityFieldQuery();
    $query->entityCondition('entity_type', 'gov_payment');
    $query->propertyCondition('uuid', $uuid);

    // Conditions.
    if (!is_null($nid)) {
      $query->propertyCondition('nid', $nid);
    }
    if (!is_null($sid)) {
      $query->propertyCondition('sid', $sid);
    }

    // Execute.
    $result = $query->execute();

    if (!empty($result['gov_payment'])) {
      return entity_load('gov_payment', array_keys($result['gov_payment']));
    }

    return [];
  }

  /**
   * Return the last error encountered.
   */
  public function getError() {
    return $this->lastError;
  }
}